﻿define(["require", "exports", "plugins/router", "durandal/app", "common/appUrl", "models/document", "models/collection", "viewmodels/deleteItems", "viewmodels/copyDocuments", "widgets/virtualTable/row", "widgets/virtualTable/column", 'models/customColumnParams', 'models/customColumns'], function(require, exports, router, app, appUrl, document, collection, deleteItems, copyDocuments, row, column, customColumnParams, customColumns) {
    var ctor = (function () {
        function ctor() {
            this.visibleRowCount = 0;
            this.recycleRows = ko.observableArray();
            this.rowHeight = 38;
            this.borderHeight = 2;
            this.viewportHeight = ko.observable(0);
            this.virtualRowCount = ko.observable(0);
            this.columns = ko.observableArray();
            this.scrollThrottleTimeoutHandle = 0;
            this.firstVisibleRow = null;
            this.itemsSourceSubscription = null;
        }
        ctor.prototype.activate = function (settings) {
            var _this = this;
            var defaults = {
                dynamicHeightTargetSelector: "footer",
                dynamicHeightBottomMargin: 0,
                selectedIndices: ko.observableArray(),
                showCheckboxes: true,
                showIds: true,
                useContextMenu: true,
                maxHeight: 'none',
                customColumnParams: {},
                isIndexMapReduce: ko.observable(true),
                isCopyAllowed: true,
                customColumns: ko.observable(customColumns.empty())
            };
            this.settings = $.extend(defaults, settings);

            if (!!settings.isIndexMapReduce) {
                this.isIndexMapReduce = settings.isIndexMapReduce;
            } else {
                this.isIndexMapReduce = ko.observable(false);
            }

            this.items = this.settings.itemsSource();
            this.focusableGridSelector = this.settings.gridSelector + " .ko-grid";
            this.virtualHeight = ko.computed(function () {
                return _this.rowHeight * _this.virtualRowCount();
            });

            this.refreshIdAndCheckboxColumn();

            this.itemsSourceSubscription = this.settings.itemsSource.subscribe(function (list) {
                _this.recycleRows().forEach(function (r) {
                    r.resetCells();
                    _this.recycleRows.valueHasMutated();
                    _this.columns.valueHasMutated();
                    r.isInUse(false);
                });
                _this.items = list;
                _this.settings.selectedIndices.removeAll();
                _this.columns.remove(function (c) {
                    return (c.binding !== 'Id' && c.binding !== '__IsChecked');
                });
                _this.gridViewport.scrollTop(0);
                _this.onGridScrolled();

                _this.refreshIdAndCheckboxColumn();
            });
        };

        ctor.prototype.attached = function () {
            var _this = this;
            this.grid = $(this.settings.gridSelector);
            if (this.grid.length !== 1) {
                throw new Error("There should be 1 " + this.settings.gridSelector + " on the page, but found " + this.grid.length.toString());
            }

            this.gridViewport = this.grid.find(".ko-grid-viewport-container");
            this.gridViewport.on('DynamicHeightSet', function () {
                return _this.onWindowHeightChanged();
            });
            this.gridViewport.scroll(function () {
                return _this.onGridScrolled();
            });

            this.setupKeyboardShortcuts();
            if (this.settings.useContextMenu) {
                this.setupContextMenu();
            }
        };

        ctor.prototype.detached = function () {
            $(this.settings.gridSelector).unbind('keydown.jwerty');

            this.gridViewport.off('DynamicHeightSet');
            if (this.itemsSourceSubscription) {
                this.itemsSourceSubscription.dispose();
            }
        };

        ctor.prototype.calculateRecycleRowCount = function () {
            var requiredRowCount = Math.ceil(this.viewportHeight() / this.rowHeight);
            var rowCountWithPadding = requiredRowCount + 10;
            return rowCountWithPadding;
        };

        ctor.prototype.createRecycleRows = function (rowCount) {
            var rows = [];
            for (var i = 0; i < rowCount; i++) {
                var newRow = new row(this.settings.showIds, this);
                newRow.createPlaceholderCells(this.columns().map(function (c) {
                    return c.binding;
                }));
                newRow.rowIndex(i);
                var desiredTop = i * this.rowHeight;
                newRow.top(desiredTop);
                rows.push(newRow);
            }

            return rows;
        };

        ctor.prototype.onGridScrolled = function () {
            var _this = this;
            this.ensureRowsCoverViewport();

            window.clearTimeout(this.scrollThrottleTimeoutHandle);
            this.scrollThrottleTimeoutHandle = setTimeout(function () {
                return _this.loadRowData();
            }, 100);
        };

        ctor.prototype.onWindowHeightChanged = function () {
            var newViewportHeight = this.gridViewport.height();
            this.viewportHeight(newViewportHeight);
            var desiredRowCount = this.calculateRecycleRowCount();
            this.recycleRows(this.createRecycleRows(desiredRowCount));
            this.ensureRowsCoverViewport();
            this.loadRowData();
        };

        ctor.prototype.setupKeyboardShortcuts = function () {
            var _this = this;
            this.setupKeyboardShortcut("DELETE", function () {
                return _this.deleteSelectedItems();
            });
            this.setupKeyboardShortcut("Ctrl+C,D", function () {
                return _this.copySelectedDocs();
            });
            this.setupKeyboardShortcut("Ctrl+C,I", function () {
                return _this.copySelectedDocIds();
            });
        };

        ctor.prototype.setupKeyboardShortcut = function (keys, handler) {
            jwerty.key(keys, function (e) {
                e.preventDefault();
                handler();
            }, this, this.settings.gridSelector);
        };

        ctor.prototype.setupContextMenu = function () {
            var _this = this;
            var untypedGrid = this.grid;
            untypedGrid.contextmenu({
                target: '#gridContextMenu',
                before: function (e) {
                    if (_this.settings.showCheckboxes == true && !_this.isIndexMapReduce()) {
                        var parentRow = $(e.target).parent(".ko-grid-row");
                        var rightClickedElement = parentRow.length ? ko.dataFor(parentRow[0]) : null;
                        if (rightClickedElement && rightClickedElement.isChecked != null && !rightClickedElement.isChecked()) {
                            _this.toggleRowChecked(rightClickedElement, e.shiftKey);
                        }
                    }
                    return true;
                }
            });
        };

        ctor.prototype.refreshIdAndCheckboxColumn = function () {
            var containsId = this.columns().first(function (x) {
                return x.binding == "Id";
            });

            if (!containsId && !this.isIndexMapReduce()) {
                if (this.settings.showCheckboxes !== false) {
                    this.columns.push(new column("__IsChecked", 38));
                }
                if (this.settings.showIds !== false) {
                    this.columns.push(new column("Id", ctor.idColumnWidth));
                }
                this.columns.valueHasMutated();
            } else if (containsId && this.isIndexMapReduce()) {
                this.columns.remove(function (c) {
                    return c.binding === 'Id' || c.binding === "__IsChecked";
                });
                this.columns.valueHasMutated();
            }
        };

        ctor.prototype.loadRowData = function () {
            var _this = this;
            if (this.items && this.firstVisibleRow) {
                var that = this;

                var firstVisibleIndex = this.firstVisibleRow.rowIndex();
                var fetchTask = this.items.fetch(firstVisibleIndex, this.recycleRows().length);
                fetchTask.done(function (resultSet) {
                    var firstVisibleRowIndexHasChanged = firstVisibleIndex !== _this.firstVisibleRow.rowIndex();
                    if (!firstVisibleRowIndexHasChanged) {
                        _this.virtualRowCount(resultSet.totalResultCount);
                        resultSet.items.forEach(function (r, i) {
                            return _this.fillRow(r, i + firstVisibleIndex);
                        });
                        _this.ensureColumnsForRows(resultSet.items);
                        _this.recycleRows.valueHasMutated();
                        _this.columns.valueHasMutated();

                        var rows = _this.recycleRows();
                        var columns = _this.columns();
                    }
                    _this.recycleRows.valueHasMutated();
                });
            }
        };

        ctor.prototype.fillRow = function (rowData, rowIndex) {
            var rowAtIndex = ko.utils.arrayFirst(this.recycleRows(), function (r) {
                return r.rowIndex() === rowIndex;
            });
            if (rowAtIndex) {
                rowAtIndex.fillCells(rowData);
                rowAtIndex.collectionClass(this.getCollectionClassFromDocument(rowData));
                rowAtIndex.editUrl(appUrl.forEditItem(rowData.getId(), appUrl.getResource(), rowIndex, this.getEntityName(rowData)));
            }
        };

        ctor.prototype.editLastSelectedItem = function () {
            var selectedItem = this.getSelectedItems(1).first();
            if (selectedItem) {
                var collectionName = this.items.collectionName;
                var itemIndex = this.settings.selectedIndices().first();
                router.navigate(appUrl.forEditItem(selectedItem.getId(), appUrl.getResource(), itemIndex, collectionName));
            }
        };

        ctor.prototype.getEntityName = function (item) {
            var obj = item;
            if (obj && obj instanceof document && obj.getEntityName) {
                var documentObj = obj;
                return documentObj.getEntityName();
            }
            return null;
        };

        ctor.prototype.getCollectionClassFromDocument = function (doc) {
            return collection.getCollectionCssClass(this.getEntityName(doc));
        };

        ctor.prototype.getColumnWidth = function (binding, defaultColumnWidth) {
            if (typeof defaultColumnWidth === "undefined") { defaultColumnWidth = 100; }
            var customColumns = this.settings.customColumns();
            var customConfig = customColumns.findConfigFor(binding);
            if (customConfig && customColumns.customMode() === true) {
                return customConfig.width();
            }

            if (binding === "Id" && defaultColumnWidth > ctor.idColumnWidth) {
                return ctor.idColumnWidth;
            }
            return defaultColumnWidth;
        };

        ctor.prototype.getColumnName = function (binding) {
            if (this.settings.customColumns().hasOverrides()) {
                var customConfig = this.settings.customColumns().findConfigFor(binding);
                if (customConfig) {
                    return customConfig.header();
                }
            }
            return binding;
        };

        ctor.prototype.ensureColumnsForRows = function (rows) {
            var columnsNeeded = {};

            if (this.settings.customColumns().hasOverrides()) {
                var colParams = this.settings.customColumns().columns();
                for (var i = 0; i < colParams.length; i++) {
                    var colParam = colParams[i];
                    columnsNeeded[colParam.binding()] = null;
                }
            } else {
                for (var i = 0; i < rows.length; i++) {
                    var currentRow = rows[i];
                    var rowProperties = currentRow.getDocumentPropertyNames();
                    for (var j = 0; j < rowProperties.length; j++) {
                        var property = rowProperties[j];
                        columnsNeeded[property] = null;
                    }
                }
            }

            for (var i = 0; i < this.columns().length; i++) {
                var colName = this.columns()[i].binding;
                delete columnsNeeded[colName];
            }

            var idColumn = this.columns.first(function (x) {
                return x.binding == "Id";
            });
            var idColumnExists = idColumn ? 1 : 0;

            var unneededColumns = new Array();
            ko.utils.arrayForEach(this.columns(), function (col) {
                if (col.binding !== "Id" && col.binding !== "__IsChecked" && rows.every(function (row) {
                    return !row.getDocumentPropertyNames().contains(col.binding);
                }))
                    unneededColumns.push(col.binding);
            });

            this.columns.remove(function (c) {
                return unneededColumns.contains(c.binding);
            });
            this.columns.valueHasMutated();
            this.settings.customColumns().columns.remove(function (c) {
                return unneededColumns.contains(c.binding());
            });

            var columnsCurrentTotalWidth = 0;
            for (var i = 2; i < this.columns().length; i++) {
                columnsCurrentTotalWidth += this.columns()[i].width();
            }

            var availiableWidth = this.grid.width() - 200 * idColumnExists - columnsCurrentTotalWidth;
            var freeWidth = availiableWidth;
            var fontSize = parseInt(this.grid.css("font-size"));
            var columnCount = 0;
            for (var binding in columnsNeeded) {
                var curColWidth = (binding.length + 2) * fontSize;
                if (freeWidth - curColWidth < 0) {
                    break;
                }
                freeWidth -= curColWidth;
                columnCount++;
            }
            var freeWidthPerColumn = (freeWidth / (columnCount + 1));

            var firstRow = this.recycleRows().length > 0 ? this.recycleRows()[0] : null;
            for (var binding in columnsNeeded) {
                var curColWidth = (binding.length + 2) * fontSize + freeWidthPerColumn;
                var columnWidth = this.getColumnWidth(binding, curColWidth);
                availiableWidth -= columnWidth;
                if (availiableWidth <= 0) {
                    break;
                }
                var columnName = this.getColumnName(binding);

                var newColumn = new column(binding, columnWidth, columnName);
                if ((binding === "Name") && (!this.settings.customColumns().customMode())) {
                    this.columns.splice(2, 0, newColumn);
                } else {
                    this.columns.push(newColumn);
                }

                var curColumnConfig = this.settings.customColumns().findConfigFor(binding);
                if (!curColumnConfig && !!firstRow) {
                    var curColumnTemplate = firstRow.getCellTemplate(binding);
                    var newCustomColumn = new customColumnParams({
                        Binding: binding,
                        Header: binding,
                        Template: curColumnTemplate,
                        DefaultWidth: availiableWidth > 0 ? Math.floor(columnWidth) : 0
                    });
                    if ((binding === "Name") && (!this.settings.customColumns().customMode())) {
                        this.settings.customColumns().columns.splice(0, 0, newCustomColumn);
                    } else {
                        this.settings.customColumns().columns.push(newCustomColumn);
                    }
                }
            }
        };

        ctor.prototype.ensureRowsCoverViewport = function () {
            var viewportTop = this.gridViewport.scrollTop();
            var viewportBottom = viewportTop + this.viewportHeight();
            var positionCheck = viewportTop;

            this.firstVisibleRow = null;
            while (positionCheck < viewportBottom) {
                var rowAtPosition = this.findRowAtY(positionCheck);
                if (!rowAtPosition) {
                    rowAtPosition = this.getOffscreenRow(viewportTop, viewportBottom);

                    var rowIndex = Math.floor(positionCheck / this.rowHeight);
                    var desiredNewRowY = rowIndex * this.rowHeight;
                    rowAtPosition.top(desiredNewRowY);
                    rowAtPosition.rowIndex(rowIndex);
                    rowAtPosition.resetCells();
                    rowAtPosition.isChecked(this.settings.selectedIndices.indexOf(rowIndex) !== -1);
                }

                if (!this.firstVisibleRow) {
                    this.firstVisibleRow = rowAtPosition;
                }

                positionCheck = rowAtPosition.top() + this.rowHeight;
            }
        };

        ctor.prototype.getOffscreenRow = function (viewportTop, viewportBottom) {
            var rows = this.recycleRows();
            for (var i = 0; i < rows.length; i++) {
                var row = rows[i];
                var rowTop = row.top();
                var rowBottom = rowTop + this.rowHeight;
                if (rowTop > viewportBottom || rowBottom < viewportTop) {
                    return row;
                }
            }

            throw new Error("Bug: couldn't find an offscreen row to recycle. viewportTop = " + viewportTop.toString() + ", viewportBottom = " + viewportBottom.toString() + ", recycle row count = " + rows.length.toString());
        };

        ctor.prototype.findRowAtY = function (y) {
            var rows = this.recycleRows();
            for (var i = 0; i < rows.length; i++) {
                var row = rows[i];
                var rowTop = row.top();
                var rowBottom = rowTop + this.rowHeight;
                if (rowTop <= y && rowBottom > y) {
                    return row;
                }
            }

            return null;
        };

        ctor.prototype.getTemplateFor = function (columnName) {
            if (this.settings.customColumns().hasOverrides()) {
                var customConfig = this.settings.customColumns().findConfigFor(columnName);
                if (customConfig) {
                    return customConfig.template();
                }
            }
            return undefined;
        };

        ctor.prototype.toggleRowChecked = function (row, isShiftSelect) {
            var _this = this;
            if (typeof isShiftSelect === "undefined") { isShiftSelect = false; }
            var rowIndex = row.rowIndex();
            var isChecked = row.isChecked();
            var firstIndex = this.settings.selectedIndices.first();
            var toggledIndices = isShiftSelect && this.settings.selectedIndices().length > 0 ? this.getRowIndicesRange(firstIndex, rowIndex) : [rowIndex];
            if (!isChecked) {
                if (this.settings.selectedIndices.indexOf(rowIndex) === -1) {
                    toggledIndices.filter(function (i) {
                        return !_this.settings.selectedIndices.contains(i);
                    }).reverse().forEach(function (i) {
                        return _this.settings.selectedIndices.unshift(i);
                    });
                }
            } else {
                this.settings.selectedIndices.removeAll(toggledIndices);
            }

            this.recycleRows().forEach(function (r) {
                return r.isChecked(_this.settings.selectedIndices().contains(r.rowIndex()));
            });
        };

        ctor.prototype.selectNone = function () {
            this.settings.selectedIndices([]);
            this.recycleRows().forEach(function (r) {
                return r.isChecked(false);
            });
        };

        ctor.prototype.selectAll = function () {
            var allIndices = [];
            for (var i = 0; i < this.items.totalResultCount(); i++) {
                allIndices.push(i);
            }
            this.settings.selectedIndices(allIndices);
            this.recycleRows().forEach(function (r) {
                return r.isChecked(true);
            });
        };

        ctor.prototype.getRowIndicesRange = function (firstRowIndex, secondRowIndex) {
            var isCountingDown = firstRowIndex > secondRowIndex;
            var indices = [];
            if (isCountingDown) {
                for (var i = firstRowIndex; i >= secondRowIndex; i--)
                    indices.unshift(i);
            } else {
                for (var i = firstRowIndex; i <= secondRowIndex; i++)
                    indices.unshift(i);
            }

            return indices;
        };

        ctor.prototype.copySelectedDocs = function () {
            this.showCopyDocDialog(false);
        };

        ctor.prototype.copySelectedDocIds = function () {
            this.showCopyDocDialog(true);
        };

        ctor.prototype.showCopyDocDialog = function (idsOnly) {
            var selectedDocs = this.getSelectedItems();
            var copyDocumentsVm = new copyDocuments(selectedDocs, this.focusableGridSelector);
            copyDocumentsVm.isCopyingDocs(idsOnly === false);
            app.showDialog(copyDocumentsVm);
        };

        ctor.prototype.getSelectedItems = function (max) {
            if (!this.items || this.settings.selectedIndices().length === 0) {
                return [];
            }
            var sliced = max ? this.settings.selectedIndices.slice(0, max) : null;
            var maxSelectedIndices = sliced || this.settings.selectedIndices();
            return this.items.getCachedItemsAt(maxSelectedIndices);
        };

        ctor.prototype.deleteSelectedItems = function () {
            var documents = this.getSelectedItems();
            var deleteDocsVm = new deleteItems(documents, this.focusableGridSelector);
            var self = this;
            deleteDocsVm.deletionTask.done(function () {
                var deletedDocIndices = documents.map(function (d) {
                    return self.items.indexOf(d);
                });
                deletedDocIndices.forEach(function (i) {
                    return self.settings.selectedIndices.remove(i);
                });
                self.recycleRows().forEach(function (r) {
                    return r.isChecked(self.settings.selectedIndices().contains(r.rowIndex()));
                });
                self.items.invalidateCache();
                self.onGridScrolled();
            });

            app.showDialog(deleteDocsVm);
        };

        ctor.prototype.getDocumentHref = function (documentId) {
            if (typeof documentId == "string") {
                return appUrl.forEditItem(documentId, appUrl.getDatabase(), null, null);
            } else {
                return "#";
            }
        };
        ctor.idColumnWidth = 200;
        return ctor;
    })();

    
    return ctor;
});
